{+--------------------------------------------------------------------------+
 | Component:   TPCharSort
 | Created:     97
 | Author:      Martin Waldenburg
 | Copyright    1997, all rights reserved.
 | Description: Sorts a Stream with PChars.
 | Version:     1.2
 | State:       FreeWare
 +--------------------------------------------------------------------------+}
 { Disclaimer:
   This is provided as is, expressly without a warranty of any kind.
   You use it at your own risc.}

unit mwPCharSort;

{$R-}


interface

uses SysUtils, Windows, Dialogs;

type

 { TPCharSort class }
  PPCharArray = ^TPCharArray;
  TPCharArray = array[0..0] of PChar;
  TPCharSortCompare = function (Item1, Item2: PChar): Integer;
  TPCharSort = class(TObject)
  private
    FSorArray, TempArray, SwapArray: PPCharArray;
    FCount: Integer;
    fCapacity:Integer;
    fOrigin:PChar;
    procedure Expand;
    procedure SetCapacity(NewCapacity:Integer);
    function GetToken(Index: Integer):PChar;
    function GetPosition(Index: Integer):Integer;
    function GetLength(Index: Integer):Integer;
    procedure Merge(PosA, PosB, PosLastA, PosLastB: LongInt; SorCompare: TPCharSortCompare);
    function Comb(jumpsize0: Integer; SorCompare: TPCHarSortCompare): boolean;
  protected
    function GetItems(Index: Integer):PChar;
    procedure SetItems(Index: Integer;  Item: PChar);
  public
    constructor create;
    destructor Destroy; override;
    function Add(Item: PChar):Integer;
    procedure Tokenize(OriginPtr: PChar; StartCapacity: Integer);
    procedure Clear;
    procedure Pack;
    procedure QuickSort(SorCompare: TPCHarSortCompare);
    procedure MergeSort(SorCompare: TPCharSortCompare);
    procedure CombSortW(SorCompare: TPCHarSortCompare);
    property Capacity:Integer read fCapacity write SetCapacity;
    property Count: Integer read FCount;
    property Items[Index: Integer]: PChar read GetItems write SetItems; default;
    property Origin:PChar read fOrigin write fOrigin;
    property Token[Index: Integer]:PChar read GetToken;
    property Position[Index: Integer]:Integer read GetPosition;
    property Length[Index: Integer]:Integer read GetLength;
  published
  end;

implementation

{ TPCharSort }

constructor TPCharSort.create;
begin
  inherited Create;
end;

procedure TPCharSort.Clear;
begin
  FCount:= 0;
  FCapacity:= 0;
  ReallocMem(FSorArray, 0 * SizeOf(PChar));
end;

destructor TPCharSort.Destroy;
begin
  Clear;
  Inherited Destroy;
end;

procedure TPCharSort.Expand;
begin
  SetCapacity(FCapacity + 1024);
end;

procedure TPCharSort.SetCapacity(NewCapacity:Integer);
begin
  FCapacity:= NewCapacity;
  ReallocMem(FSorArray, FCapacity * SizeOf(PChar));
end;

function TPCharSort.Add(Item: PChar):Integer;
begin
  Result := FCount;
  if Result = FCapacity then Expand;
  FSorArray[Result] := Item;
  Inc(FCount);
end;

function TPCharSort.GetItems(Index: Integer):PChar;
begin
  Result := FSorArray[Index];
end;

procedure TPCharSort.SetItems(Index: Integer;  Item: PChar);
begin
  FSorArray[Index]:= Item;
end;

procedure TPCharSort.Pack;
begin
  ReallocMem(FSorArray, FCount * SizeOf(PChar));
end;

function TPCharSort.GetLength(Index: Integer):Integer;
var
  First, Final: PChar;
begin
  First:= FSorArray[Index];
  Final:= First;
  while Final^ <> #13 do inc(Final);
  Result:= Final - First;
end;

function TPCharSort.GetPosition(Index: Integer):Integer;
var
  First: PChar;
begin
  First:= FSorArray[Index];
  Result:= First - FOrigin;
end;

function TPCharSort.GetToken(Index: Integer):PChar;
var
  First: PChar;
  TokenLen: Integer;
  aToken: array[0..4096] of Char;
begin
  First:= FSorArray[Index];
  TokenLen:= GetLength(Index);
  StrLCopy(aToken, First, TokenLen);
  aToken[TokenLen]:= #0;
  Result:= aToken;
end;

procedure TPCharSort.Tokenize(OriginPtr: PChar; StartCapacity: Integer);
var
  Start, Run: PChar;
begin
  fOrigin:= OriginPtr;
  SetCapacity(StartCapacity);
  Run:= FOrigin;
  while (Run^ > #0) do
  begin
    Start:= Run;
    Add(Start);
    while (Run^ > #0) and (Run^ <> #13) do Inc(Run);
    while (Run^ = #10) or (Run^ = #13) do Inc(Run);
  end;
end;

{ Based on a non-recursive QuickSort from the SWAG-Archive.
  ( TV Sorting Unit by Brad Williams ) }
procedure TPCharSort.QuickSort(SorCompare: TPCharSortCompare);
var Left, Right, SubArray, SubLeft, SubRight:LongInt;
    Temp, Pivot: Pointer;
    Stack : array[1..32] of record First, Last : LongInt; end;
begin
  SubArray := 1;
  Stack[SubArray].First := 0;
  Stack[SubArray].Last := Count - 1;
    repeat
      Left := Stack[SubArray].First;
      Right := Stack[SubArray].Last;
      Dec(SubArray);
      repeat
        SubLeft := Left;
        SubRight := Right;
        Pivot := FSorArray[(Left + Right) shr 1];
        repeat
          while SorCompare(FSorArray[SubLeft], Pivot) < 0 do Inc(SubLeft);
          while SorCompare(FSorArray[SubRight], Pivot) > 0 do Dec(SubRight);
          IF SubLeft <= SubRight then
            begin
              Temp := FSorArray[SubLeft];
              FSorArray[SubLeft] := FSorArray[SubRight];
              FSorArray[SubRight] := Temp;
              Inc(SubLeft);
              Dec(SubRight);
            end;
        until SubLeft > SubRight;
        IF SubLeft < Right then
          begin
            Inc(SubArray);
            Stack[SubArray].First := SubLeft;
            Stack[SubArray].Last := Right;
          end;
          Right := SubRight;
      until Left >= Right;
    until SubArray = 0;
end;

{Unfortunately the " Merge " routine needs additional memory
 An Algorithm to perform merging in linear time without extra space
 is described in:
 B. Huang and M. Langston, " Practical In-place Merging ",
 Communications of the ACM 31(1988), 348-352. }
procedure TPCharSort.Merge(PosA, PosB, PosLastA, PosLastB: LongInt; SorCompare: TPCharSortCompare);
var
  TempPos : integer;
begin
  TempPos := PosA;
  while ( PosA <= PosLastA ) and ( PosB <= PosLastB ) do  { Main Loop }
    begin
      if SorCompare(FSorArray[PosA], FSorArray[PosB]) <= 0 then
        begin
          TempArray[ TempPos ] := FSorArray[ PosA ];
          inc(PosA);
        end
      else
        begin
          TempArray[ TempPos ] := FSorArray[ PosB ];
          inc(PosB);
        end;
          inc(TempPos);
    end;

  while PosA <= PosLastA do    { Copy Rest of First Half }
    begin
      TempArray[ TempPos ] := FSorArray[ PosA ];
      inc(TempPos); inc(PosA);
    end;

  while PosB <= PosLastB do   { Copy Rest of Second Half }
    begin
      TempArray[ TempPos ] := FSorArray[ PosB ];
      inc(TempPos); inc(PosB);
    end;
end;

{Non-recursive Mergesort.
 Very fast, if enough memory available.
 The number of comparisions used is nearly optimal, about 3/4 of QuickSort.
 If comparision plays a very more important role than exchangement,
 it outperforms QuickSort in any case.
 ( Large keys or a time intensive comparision like AnsiCompareText )
 From all Algoritms with O(N lg N) it's the only stable, meaning it lefts
 equal keys in the order of input. This may be important in some cases.}
procedure TPCharSort.MergeSort(SorCompare: TPCharSortCompare);
var
  a, b, N, todo: LongInt;
begin
  ReallocMem(TempArray, (FCount) * SizeOf(Pointer));
  N:= 1;
  repeat
      todo:= 0;
      repeat
        a:= todo;
        b:= a +N;
        todo:= b +N;
        if b > FCount -1 then b:= FCount -1;
        if todo > FCount then todo:= FCount;
        Merge(a, b, b -1, todo -1, SorCompare);
      until todo = Fcount;
      SwapArray:= FSorArray; {Alternating use of the arrays.}
      FSorArray:= TempArray;
      TempArray:= SwapArray;
      N:= N shl 1;
    until N > Fcount;
    ReallocMem(TempArray, 0);
end;

{ Multipication by 0.76 gives a slightly better result than
  division by 1.3.
  Because of the FOR loop it runs faster on arrays starting with one}
function TPCharSort.Comb(jumpsize0: Integer; SorCompare: TPCharSortCompare): boolean;
const irgap: Real = 0.76;
var
  i, SwapNumber: Integer;
  Temp: Pointer;
begin
 repeat
    jumpsize0 := trunc(jumpsize0*irgap);
    SwapNumber:= jumpsize0+1;
    Result:= True;
    for i := 0 to (FCount-jumpsize0-2) do
      begin
        if SorCompare(FSorArray[i], FSorArray[SwapNumber]) > 0 then
          begin
            Temp := FSorArray[i];
            FSorArray[i] := FSorArray[SwapNumber];
            FSorArray[SwapNumber] := Temp;
            Result := false;
          end;
          inc(SwapNumber);
      end;
  until jumpsize0 = 0;
end;

{Driver for the " Comb " routine.
 Based on routines from the SWAG-Archive.
 Very fast, for a smaller number of items with large keys
 " Comb " may outperform Quicksort.
 ( Only a few thousends )}
procedure TPCharSort.CombSortW(SorCompare: TPCharSortCompare);
begin
  comb(FCount, SorCompare);
  repeat
  until comb(0, SorCompare);
end;

end.
