{+--------------------------------------------------------------------------+
 | Component:   TPStringSort
 | Created:     97
 | Author:      Martin Waldenburg
 | Copyright    1997, all rights reserved.
 | Description: Fast in memory sorter for Strings.
 | Version:     1.2
 | State:       FreeWare
 +--------------------------------------------------------------------------+}
 { Disclaimer:
   This is provided as is, expressly without a warranty of any kind.
   You use it at your own risc.}

unit mwPStringSort;

{$R-}


interface

uses SysUtils, Windows, Dialogs;

type

 { TPStringSort class }
  PString = ^TaString;
  TaString = record Data: String end;
  PStringArray = ^TStringArray;
  TStringArray = array[0..0] of PString;
  TPStringSortCompare = function (Item1, Item2: PString): Integer;
  TPStringSort = class(TObject)
  private
    FSorArray, TempArray, SwapArray: PStringArray;
    FCount: Integer;
    fCapacity:Integer;
    fOrigin:PChar;
    FString: PString;
    procedure Expand;
    procedure SetCapacity(NewCapacity:Integer);
    function GetToken(Index: Integer):String;
    procedure Merge(PosA, PosB, PosLastA, PosLastB: LongInt; SorCompare: TPStringSortCompare);
    function Comb(jumpsize0: Integer; SorCompare: TPStringSortCompare): boolean;
  protected
    function GetItems(Index: Integer):PString;
    procedure SetItems(Index: Integer;  Item: PString);
  public
    constructor create;
    destructor Destroy; override;
    function Add(Item: String):Integer;
    procedure Tokenize(OriginPtr: PChar; StartCapacity: Integer);
    procedure Clear;
    procedure Pack;
    procedure QuickSort(SorCompare: TPStringSortCompare);
    procedure MergeSort(SorCompare: TPStringSortCompare);
    procedure CombSortW(SorCompare: TPStringSortCompare);
    property Capacity:Integer read fCapacity write SetCapacity;
    property Count: Integer read FCount;
    property Items[Index: Integer]: PString read GetItems write SetItems; default;
    property Origin:PChar read fOrigin write fOrigin;
    property Token[Index: Integer]:String read GetToken;
  published
  end;

implementation

{ TPStringSort }

constructor TPStringSort.create;
begin
  inherited Create;
end;

procedure TPStringSort.Clear;
var
  I: Integer;
begin
  for I:= 0 to Count -1 do
  begin
    FString:= FSorArray[I];
    Dispose(FString);
  end;
  FCount:= 0;
  FCapacity:= 0;
  ReallocMem(FSorArray, 0 * SizeOf(PString));
end;

destructor TPStringSort.Destroy;
begin
  Clear;
  Inherited Destroy;
end;

procedure TPStringSort.Expand;
begin
  SetCapacity(FCapacity + 1024);
end;

procedure TPStringSort.SetCapacity(NewCapacity:Integer);
begin
  FCapacity:= NewCapacity;
  ReallocMem(FSorArray, FCapacity * SizeOf(PString));
end;

function TPStringSort.Add(Item: String):Integer;
begin
  Result := FCount;
  if Result = FCapacity then Expand;
  New(FString);
  FString^.Data:= Item;
  FSorArray[Result] := FString;
  Inc(FCount);
end;

function TPStringSort.GetItems(Index: Integer):PString;
begin
  Result := FSorArray[Index];
end;

procedure TPStringSort.SetItems(Index: Integer;  Item: PString);
begin
  FSorArray[Index]:= Item;
end;

procedure TPStringSort.Pack;
begin
  ReallocMem(FSorArray, FCount * SizeOf(PString));
end;

function TPStringSort.GetToken(Index: Integer):String;
begin
  FString:= FSorArray[Index];
  Result := FString^.Data;
end;

procedure TPStringSort.Tokenize(OriginPtr: PChar; StartCapacity: Integer);
var
  TokenLen: Integer;
  Start, Run: PChar;
  aToken: array[0..4095] of Char;
begin
  fOrigin:= OriginPtr;
  SetCapacity(StartCapacity);
  Run:= FOrigin;
  while (Run^ > #0) do
  begin
    Start:= Run;
    while (Run^ > #0) and (Run^ <> #13) do Inc(Run);
    TokenLen:= Run - Start;
    StrLCopy(aToken, Start, TokenLen);
    aToken[TokenLen]:= #0;
    Add(aToken);
    while (Run^ = #10) or (Run^ = #13) do Inc(Run);
  end;
end;

{ Based on a non-recursive QuickSort from the SWAG-Archive.
  ( TV Sorting Unit by Brad Williams ) }
procedure TPStringSort.QuickSort(SorCompare: TPStringSortCompare);
var Left, Right, SubArray, SubLeft, SubRight:LongInt;
    Temp, Pivot: Pointer;
    Stack : array[1..32] of record First, Last : LongInt; end;
begin
  SubArray := 1;
  Stack[SubArray].First := 0;
  Stack[SubArray].Last := Count - 1;
    repeat
      Left := Stack[SubArray].First;
      Right := Stack[SubArray].Last;
      Dec(SubArray);
      repeat
        SubLeft := Left;
        SubRight := Right;
        Pivot := FSorArray[(Left + Right) shr 1];
        repeat
          while SorCompare(FSorArray[SubLeft], Pivot) < 0 do Inc(SubLeft);
          while SorCompare(FSorArray[SubRight], Pivot) > 0 do Dec(SubRight);
          IF SubLeft <= SubRight then
            begin
              Temp := FSorArray[SubLeft];
              FSorArray[SubLeft] := FSorArray[SubRight];
              FSorArray[SubRight] := Temp;
              Inc(SubLeft);
              Dec(SubRight);
            end;
        until SubLeft > SubRight;
        IF SubLeft < Right then
          begin
            Inc(SubArray);
            Stack[SubArray].First := SubLeft;
            Stack[SubArray].Last := Right;
          end;
          Right := SubRight;
      until Left >= Right;
    until SubArray = 0;
end;

{Unfortunately the " Merge " routine needs additional memory
 An Algorithm to perform merging in linear time without extra space
 is described in:
 B. Huang and M. Langston, " Practical In-place Merging ",
 Communications of the ACM 31(1988), 348-352.}
procedure TPStringSort.Merge(PosA, PosB, PosLastA, PosLastB: LongInt; SorCompare: TPStringSortCompare);
var
  TempPos : integer;
begin
  TempPos := PosA;
  while ( PosA <= PosLastA ) and ( PosB <= PosLastB ) do  { Main Loop }
    begin
      if SorCompare(FSorArray[PosA], FSorArray[PosB]) <= 0 then
        begin
          TempArray[ TempPos ] := FSorArray[ PosA ];
          inc(PosA);
        end
      else
        begin
          TempArray[ TempPos ] := FSorArray[ PosB ];
          inc(PosB);
        end;
          inc(TempPos);
    end;

  while PosA <= PosLastA do    { Copy Rest of First Half }
    begin
      TempArray[ TempPos ] := FSorArray[ PosA ];
      inc(TempPos); inc(PosA);
    end;

  while PosB <= PosLastB do   { Copy Rest of Second Half }
    begin
      TempArray[ TempPos ] := FSorArray[ PosB ];
      inc(TempPos); inc(PosB);
    end;
end;

{Non-recursive Mergesort.
 Very fast, if enough memory available.
 The number of comparisions used is nearly optimal, about 3/4 of QuickSort.
 If comparision plays a very more important role than exchangement,
 it outperforms QuickSort in any case.
 ( Large keys or a time intensive comparision like AnsiCompareText )
 From all Algoritms with O(N lg N) it's the only stable, meaning it lefts
 equal keys in the order of input. This may be important in some cases.}
procedure TPStringSort.MergeSort(SorCompare: TPStringSortCompare);
var
  a, b, N, todo: LongInt;
begin
  ReallocMem(TempArray, (FCount) * SizeOf(Pointer));
  N:= 1;
  repeat
      todo:= 0;
      repeat
        a:= todo;
        b:= a +N;
        todo:= b +N;
        if b > FCount -1 then b:= FCount -1;
        if todo > FCount then todo:= FCount;
        Merge(a, b, b -1, todo -1, SorCompare);
      until todo = Fcount;
      SwapArray:= FSorArray; {Alternating use of the arrays.}
      FSorArray:= TempArray;
      TempArray:= SwapArray;
      N:= N shl 1;
    until N > Fcount;
    ReallocMem(TempArray, 0);
end;

{ Multipication by 0.76 gives a slightly better result than
  division by 1.3.
  Because of the FOR loop it runs faster on arrays starting with one}
function TPStringSort.Comb(jumpsize0: Integer; SorCompare: TPStringSortCompare): boolean;
const irgap: Real = 0.76;
var
  i, SwapNumber: Integer;
  Temp: Pointer;
begin
 repeat
    jumpsize0 := trunc(jumpsize0*irgap);
    SwapNumber:= jumpsize0+1;
    Result:= True;
    for i := 0 to (FCount-jumpsize0-2) do
      begin
        if SorCompare(FSorArray[i], FSorArray[SwapNumber]) > 0 then
          begin
            Temp := FSorArray[i];
            FSorArray[i] := FSorArray[SwapNumber];
            FSorArray[SwapNumber] := Temp;
            Result := false;
          end;
          inc(SwapNumber);
      end;
  until jumpsize0 = 0;
end;

{Driver for the " Comb " routine.
 Based on routines from the SWAG-Archive.
 Very fast, for a smaller number of items with large keys
 " Comb " may outperform Quicksort.
 ( Only a few thousends )}
procedure TPStringSort.CombSortW(SorCompare: TPStringSortCompare);
begin
  comb(FCount, SorCompare);
  repeat
  until comb(0, SorCompare);
end;

end.
